##############################################################################
#
#  List the test cases that are to be disabled temporarily.
#
#  Separate the test case name and the comment with ':'.
#
#    <testcasename> : BUG#<xxxx> <date disabled> <disabler> <comment>
#
#  Do not use any TAB characters for whitespace.
#
##############################################################################
transaction_prealloc_size_basic_32  : Bug#11748572
transaction_prealloc_size_basic_64  : Bug#11748572
query_cache_size_basic_32           : Bug#13535584
query_cache_size_basic_64           : Bug#11748572
#thread_cache_size_func             : Bug#11750172: 2008-11-07 joro main.thread_cache_size_func fails in pushbuild when run with pool of threads

