##############################################################################
#
#  List the test cases that are to be disabled temporarily.
#
#  Separate the test case name and the comment with ':'.
#
#    <testcasename> : BUG#<xxxx> <date disabled> <disabler> <comment>
#
#  Do not use any TAB characters for whitespace.
#
##############################################################################
lowercase_table3         : Bug#11762269 2010-06-30 alik main.lowercase_table3 on Mac OSX
read_many_rows_innodb    : Bug#11748886 2010-11-15 mattiasj report already exists
sum_distinct-big         : Bug#11764126 2010-11-15 mattiasj was not tested
archive-big              : Bug#11817185 2011-03-10 Anitha Disabled since this leads to timeout on Solaris Sparc
log_tables-big           : Bug#11756699 2010-11-15 mattiasj report already exists
mysql_embedded           : Bug#12561297 2011-05-14 Anitha Dependent on PB2 changes - eventum#41836
